//**********************************************************
// Name : main.c (Intel-Hex-Converter/EPS1-loader)
// Fktn.: HEX-Record >> EPS1 Protocol (Win32 Console App) 
// Contr: Philips 8051 ELV9996 EPS1 (ELV EpromSimulator)
// Appl.: load EPS1 via V24 (RS232)
// Autor: GG
// Firma: EAST ;) , 13.08.2013 (C) 2014
// Doku : EPR_SIM.HLP\Datenbertragungsformat (gehrt zur Win32 Software),
//        Bauanleitung EPS1_KM_G_020115.pdf
//        ELVjournal 2/99 (laut PDF-> weie Schrift auf weiem Hintergrund)
//        ELVjournal 4/99 (Teil 3 und RAM-Simulator-Zusatzplatinen 6264/62256/62100)
// Changes:
//        Emulation 2716 (fester Offset 0x2800)
//        wake up EPS1 via DTR 100ms low & >100ms high
//**********************************************************
#define  _CRT_SECURE_NO_WARNINGS // fopen >> fopen_s
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <windows.h>


#define MAX_MEMORY   0x10000L
#define RAM_END     (MAX_MEMORY-1)

//*******************************************************************************

// EPS1 Hardware
// pin1 & pin31 n.c.
// pin30    & pin32 5V
// pin2 A16 & pin3 A15 jeweils mit pull-down 47k onboard

// Adapter fr 2716  >>>>  offset 0x2800
// pin25 A11 5V (2732.A11 >> 2716.PGM 5V)
// pin28 A13 5V >> pin30
// pin4 A12 & pin29 A14 gemeinsamer 33k pull-down

//*******************************************************************************
#define offs 0x2800  // lade an Adresse 0x2800 (A13 High & A11 High)

typedef unsigned char byte;
typedef unsigned char word[2]; //unsigned short
typedef unsigned char ulong[4];


//*******************************************************************************

// Protokoll: 8bit ASCII mit code shifting (+16) fr SOH/EOT/DLE (beim Senden)
//            bzw. STX/ETX/DLE (beim Empfang) per DLE (data link escape) einfgen
// Hardware-
// parameter: 9600bd (@start), odd parity, one stop bit, no hardware flow control
//            (DTR set -> device reset; DTR clr; wait 200ms; max. 1s to connect)
//*******************************************************************************
      // NUL 0
#define  SOH  1  // start of header
#define  STX  2  // start of text
#define  ETX  3  // (end  of text?)
#define  EOT  4  // (end  of transmission?)
      // ENQ  5
#define  ACK  6  // acknowledge
      // BEL  7
      // BS   8
      // HT   9
      // LF   10
      // VT   11 (verTab)
      // FF   12
      // CR   13
      // SO   14 (shiftout?)
      // SI   15 (shiftin?)
#define  DLE  16
#define  DC1  17  // SOH+DLE
#define  DC2  18  // STX+DLE
#define  DC3  19  // ETX+DLE
#define  DC4  20  // EOT+DLE
#define  NAK  21  // no acknowledge
      // ESC  27
#define SPACE 32  // DLE+DLE

// PC to EPS1:      <SOH>,[Daten],<CRCH/L>,<EOT>             // CRC 16bit (X16+X12+X5+1) ber Header(SOH,STX/ACK/NAK) und Daten ohne DLEs
// EPS1 response:   <STX>,<ACK>,[response],<CRCH/L>,<ETX>
//                  <STX>,<NAK>,[errornum],<CRCH/L>,<ETX>    // errornum: '0' parity error; '1' illegal/unknown cmd; '2' buffer overrun;
                                                             //           '3' CRC16 error;  '4' parameter error

// PC commands
#define CMD_WBLK0   '0' // <AdrLow>,<AdrHigh>,<64Byte data> . . write block 0 (0..FFFF)
#define CMD_WBLK1   '1' // <AdrLow>,<AdrHigh>,<64Byte data>     write block 1 (10000..1FFFF)
#define CMD_WBLK32  '2' // <AdrLow>,<AdrHigh>,<64Byte data> . . write both blocks (copy block 0..7FFF to 8000..FFFF)
#define CMD_RBLK0   '8' // <AdrLow>,<AdrHigh>,<bsize <= 64>     read block 0
#define CMD_RBLK1   '9' // <AdrLow>,<AdrHigh>,<bsize <= 64> . . read block 1
#define CMD_BAUDR   'B' // <'0'..'6'>                           set baud 4800,9600,14400,19200,28800,38400,57600 (ACK before BR change)
#define CMD_RESET   'R' // <'0'..'3'> . . . . . . . . . . . . . reset    none,res,/res,both active
#define CMD_TIMEOUT 'T' // <0,1..255>                           timeout  off, x*20ms
#define CMD_PWRDWN  'P' //  . . . . . . . . . . . . . . . . . . power down (reset off)
#define CMD_INIT    'I' // <byte>                               init RAM memset(0,0x1FFFF,byte) sends ACK after Init complete (ca. 100ms delayed)

const unsigned char cmdTimeoutOff[6]={SOH,CMD_TIMEOUT,0,0xF5,0x4B,EOT};
const unsigned char cmdPowerDown[5]={SOH,CMD_PWRDWN,0x69,0xC4,EOT};
const unsigned char cmdResetOn[6]={SOH,CMD_RESET,'3',0x59,0xDD,EOT};
const unsigned char cmdFillFF[6]={SOH,CMD_INIT,255,0x9E,0x94,EOT};
const unsigned char cmdFill0[6]={SOH,CMD_INIT,0,0x80,0x64,EOT};

/* EPS1 Response
const unsigned char noErrorResponse[5]={STX,ACK,6,0xA4,ETX};     // Antwort no Error
const unsigned char errParamWrng[6]={STX,NAK,'4',0xE4,0x31,ETX}; // Antwort Param not valid error or too many params (Reset '4'..)
const unsigned char errCRCfailed[6]={STX,NAK,'3',0x94,0xD6,ETX}; // Antwort CRC error
const unsigned char errBufferOvr[6]={STX,NAK,'2',0x84,0xF7,ETX}; // Antwort buffer overrun error (inkl. SOH & EOT max. 71 Zeichen ohne DLEs)
const unsigned char errNoCommand[6]={STX,NAK,'1',0xB4,0x94,ETX}; // Antwort illegal command error
const unsigned char errParityErr[6]={STX,NAK,'0',0xA4,0xB5,ETX}; // Antwort Parity error
*/

//*******************************************************************************
//*******************************************************************************


HANDLE hPort;                        // COM
DWORD comResult;
COMMTIMEOUTS comTimeout={4, 2, 70, 0, 0}; // 4ms zwischen Zeichen; n*2ms + 70ms
DCB  DeviceCon;
char * comInterface;

FILE * input;

char LineBuffer[200];                // 64 Bytes / HEX-Record
short int BufferSrc[MAX_MEMORY];     // RAM (1 Word pro Byte)

unsigned char byteHRecChecksumm;     // HEX-Record Checksumme fr eine Zeile


//*******************************************************************************

//    CRC16   Prfsumme

//*******************************************************************************
/*
 * this source is part of the linux kernel distribution (crc16.c from sdcc)
 * Copyright is held by AT&T.
 *	Calculate a 16-bit CRC checksum using ITU-T V.41 polynomial.
 *	The OSTA-UDF(tm) 1.50 standard states that using CRCs is mandatory.
 *	The polynomial used is:	x^16 + x^12 + x^5 + 1
 */

unsigned crc_table[256];

unsigned crc16(const unsigned char *data, unsigned size) // funktioniert
 {
  unsigned crc=0;

  if(size) do
    crc = crc_table[(crc >> 8 ^ *(data++)) & 0xffU] ^ (crc << 8);
  while (--size);
  return (unsigned short)crc;
 }

void genCrcTable(unsigned poly) // funktioniert (Tabelle ist richtig)
 {
  unsigned i, n, crc;

  for (n = 0; n < 256; n++)
   {
    crc = n << 8;
    for (i = 0; i < 8; i++)
     {
      if(crc & 0x8000U) crc = (crc << 1) ^ poly;
      else            crc <<= 1;
      crc &= 0xFFFFU;
     }
    crc_table[n]=crc;  // _cprintf("error %d %x!=%x\r\n", n, crc_table[n], crc);
   }
 }
//*******************************************************************************
// end of "linux kernel" (code from sdcc)
//*******************************************************************************




//*******************************************************************************
// sende eine Zeichenkette (&cmd, sizeof(cmd)) -> return 0 on error
//  und
// lese zurckgegebene Antwort in (LineBuffer)
//      Auswertung Antwort auf cmd '8' & '9': LineBuffer[] verarbeitete Daten
//                                            return       sizeof(Daten)
//      Auswertung Antwort auf cmd '0'-'2'/'B'/'I'/'P'/'R'/'T':
//                                            Antwort ACK -> return n; (len cmd)
//                                            Antwort NAK -> return 0;
//*******************************************************************************

int writeCmd(const unsigned char* cmd, unsigned n)
 {
  unsigned i=3, j, m; // retry

  do
   {
    WriteFile(hPort, cmd, n, &comResult, NULL);
    FlushFileBuffers(hPort); // only the buffer of device driver
    if(comResult!=n) {_cprintf("error write COM\n"); return 0;}

    j=3;m=0;
    do
     {
      ReadFile(hPort,&LineBuffer[m], sizeof(LineBuffer)-m, &comResult, NULL);   // wartet max. 74ms
      if(comResult>0)
       {
        if((m+=comResult)>sizeof(LineBuffer)) {_cprintf("error comResult %x\n", comResult); return 0;}
        if(m>4) if(LineBuffer[m-1]==(char)ETX) goto checkResult;
       }
      Sleep(10); // unntig, strt aber auch nicht (wegen des eventChars sollten eigentlich immer alle bytes auf einmal gelesen werden)
     }while(--j);

    if(m<5) goto resetRxBuf;
    if(LineBuffer[m-1]!=(char)ETX) goto resetRxBuf;
checkResult:
//////////////////////// Datenblock lesen ////////////////////////
    if((*(cmd+1)&(unsigned char)0xFE) == '8')    // cmd '8' & '9' read block
     {
      unsigned char c;
      unsigned pos;

      j=m-1; // skip ETX
      do
       {
        if((c=LineBuffer[--j])==ETX) goto error; 
        if(c==STX) goto foundStart;
       }while(j);
      goto error;

foundStart:
      pos=0;
      m-=j+1;
      if(m<5)                        goto error; // STX+ACK+data+CRCH+CRCL(+ETX)

      do
       {
        c=LineBuffer[j++];
        if(c==DLE)
         {
          if(!--m)                   goto error; // no more data
          if((c=LineBuffer[j++])<16) goto error;
          c-=16;
         }
        LineBuffer[pos++]=c;
       }while(--m);

      m=(unsigned char)LineBuffer[pos-1];        // CRCL
      m+=(unsigned)LineBuffer[pos-=2]<<8;        // CRCH       (pos -= 2; !!!!)

      if(crc16((unsigned char*)&LineBuffer[0], pos)==m)  return pos;
      goto error;
     }
    else
//////////////////////// Antwort Datenblock schreiben ('0'-'2') oder cmd 'B/I/P/R/T' ////////////////////////
    if(*(unsigned long*)&LineBuffer[m-5] == (0xA4060000L + (unsigned)ACK*256+STX))  return n;
    
//  if(LineBuffer[m-5]==NAK && LineBuffer[m-4]!='3') printHex((unsigned char*)&LineBuffer[0], m);
    if(m>5) if(*(unsigned short*)&LineBuffer[m-6] == STX + NAK*256) _cprintf("%c",LineBuffer[m-4]);
error:
    Sleep(10); // Synchronize after error
resetRxBuf:
    PurgeComm( hPort, /*PURGE_TXABORT |*/ PURGE_RXABORT | /*PURGE_TXCLEAR |*/ PURGE_RXCLEAR ) ;

   }while(--i); // retry
  return 0;     // error
 }


//*******************************************************************************
// sende Datenblock s(SOH, cmd, addr.l, addr.h, data 1..n, CRCH, CRCL), sizeof(s)
//*******************************************************************************

int sendBlock( char* s, unsigned m ) // Puffer 's' enthlt bereits SOH weil CRC16(s,m-2) alles auer CRCH/L und EOT bergeben bekommt
 {
  unsigned n=2;
  unsigned char c;
  unsigned char line[128];

  if(m<4 || m>64+6 || *s++!=SOH) return 0;
  line[0] = SOH;
  line[1] = *s++; // cmd '0' .. '2'
  m-=2;

  do
   {
    c=*s++;
    if(c==DLE||c==EOT||c==SOH) {line[n++]=DLE;c+=16;} // c + DLE
    line[n++]=c;
    if(!--m) break;
   } while (n<sizeof(line)-2); // Platz fr DLE+'c'+EOT

  line[n++]=EOT; // EOT wird erst hier angefgt und ist nicht in der Byte-Anzahl m enthalten

  if(writeCmd(line, n)==n) {_cprintf("."); return n;} // ok
                            _cprintf("!"); return 0;  // error
 }


//////////////// hexadezimale Ziffer nach binr ////////////

// Eingabe:  ASCII Zeichen '0'..'9'; 'a'..'f'; 'A'..'F'

// Ergebnis: 0..15 entsprechend HEX-Ziffer
//              -1 alle anderen Zeichen 

////////////////////////////////////////////////////////////

char hexDigit2bin(register char c)
 {
  if(c>='A')                         //          A........f
   {
    if(c>='a')                       //
     {
      if(c<='f') return c-('a'-10);  //                a..f
      else       return -1;          // >f
     }
    if(c<='F')   return c-('A'-10);  //          A..F
    else         return -1;          // >F <a
   }
  if(c>='0')                         //    0.....A
   {
    if(c<='9')   return c-'0';       //    0..9
   }
  return -1;                         // <0 >9 <A
 }

/////////////// hexadezimal nach binr ///////////////////

// Eingabe:
//   line : Zeiger auf Zeichenkette ('0'..'9', 'A'..'F', 'a'..'f')
//          ohne Leerzeichen oder terminierende Null
//   size : Anzahl der Digits in Zeichenkette

// Ergebnis:
//   function result:
//           positive Zahl bei size(<8) gltigen Zeichen
//           -1 wenn Fehler
//   global byteHRecChecksumm:
//           wird erhht entsprechend Inhalt der gelesenen Bytes
//           (bei ungerader Anzahl Digits Addition rechtsbndig
//           bei result -1: Checksumme nicht gltig)

//////////////////////////////////////////////////////////

long hex2bin(char * line, char size)
 {
  long temp=0;

  while(size-- > 0)
   {
    temp=(temp<<4) | (long)hexDigit2bin(*(line++));

    if (!(size&1)) byteHRecChecksumm+=(unsigned char)temp;
   }
  return temp;
 }

void printHex(unsigned char* s, unsigned n)
 {
  if(n)
   {
    if(--n) do _cprintf(" %02x,", *s++); while(--n);
    _cprintf(" %02x\r\n", *s);
   }
 }

//////////////////////////////////////////////////////////

//  M A I N

//  Intel HEX File lesen; COM-Port ffnen; Senden

//  Eingabe:
//       HEX Records beliebiger Gre und Reihenfolge

//  Verarbeitung:
//       Binrumwandlung

//  Ausgabe:
//       an EPS1 in 2716 Konfiguration senden (mit Offset 
//                  der Adressen 0x2800)
//       EPS1 in den power-down schalten (Reset freigeben)

//////////////////////////////////////////////////////////

void main(int argc, char *argv[], char * envp[] )
{
 long Adresse;
 unsigned int error, i, j, lineNr, byteCnt;
 unsigned int temp;

 error=0;
 lineNr=0;

 _cprintf("EPS1 loader V1.0a\n");
 _cprintf("programming the EPS1(8051) Memory with Intel Hex Files\n");
 _cprintf("Author:GG, EAS C 2013\n");
 genCrcTable(0x1021);

 if(argc!=3)
  {
   _cprintf("Fehler Parameter: <COMx:> <Src>\n");
   i=40;
   while(i--&&!_kbhit()) Sleep(100); // Verzgerung Console Schlieen bei Start von Desktop
   while(_kbhit()) _getch();         // Keys einlesen (fr manche Tasten mssen zwei Byte eingelesen werden)
   return;
  }

/* if(argc>3)
  {
   sscanf(argv[3], "%x", &offs)
  }*/

 memset(BufferSrc, -1, sizeof(BufferSrc));

 comInterface=argv[1];
 input = fopen(argv[2], "rt"); //text
 if (!input)
  {
   _cprintf("Fehler Datei oeffnen: %s\n", argv[2]);
   return;
  }
 else
  {
   lineNr=0;
   while(!feof(input))
    {
     if(!fgets(LineBuffer, sizeof(LineBuffer), input)) break;
     lineNr++;

     if(LineBuffer[0]&&(LineBuffer[0]!=13)) // Leerzeilen
      {
       if(LineBuffer[0]!=':')
        {
         _cprintf("Error in Intel HEX-Record Zeile:%d ':' erwartet!\n", lineNr);
         break;
        }
       else
        {
         error=0;
         byteHRecChecksumm=0;
         byteCnt=hex2bin(&LineBuffer[1], 2);
         Adresse=hex2bin(&LineBuffer[3], 4);

         if(error) goto illegalRecord;

         j=9;

         switch(hex2bin(&LineBuffer[7], 2))
          {
           case 00:
             if(((unsigned long)Adresse>=MAX_MEMORY)||(Adresse+byteCnt>MAX_MEMORY)) error=-1;
             break;
           case 01:
             /*if(startAddress!=-1)
              {
               _cprintf("startAddress twice at:%d!\n", lineNr);
               continue;
              }

             startAddress=Adresse;*/

             if(!byteCnt) goto checkLastRecord;
           default:
illegalRecord:
             _cprintf("illegal HEX-Record Typ Zeile:%d!\n", lineNr);
             continue;    // next line
          }

         if((unsigned)byteCnt>64)
          {
           if(byteCnt>255) _cprintf("Zeichenfehler size Byte Zeile:%d!\n", lineNr);
           else            _cprintf("size Byte zu gro Zeile:%d!\n", lineNr);
           break;
          }

         if(error)
          {
           _cprintf("Fehler im Adressfeld Zeile:%d!\n", lineNr);
           break;
          }


      ////////////// bytes lesen ///////////////
         for(i=0; i<byteCnt; i++)
          {
           if(j>(sizeof(LineBuffer)-2)) break;
           if((temp=hex2bin(&LineBuffer[j], 2))<0) error|=4;

           if(BufferSrc[Adresse+i]>=0) error|=1;
           BufferSrc[Adresse+i]=temp;
           j+=2;
          }

      /////////////// testen ///////////////////

         if(error)
          {
           if((unsigned)error<3)
             _cprintf("Fehler %d doppelte HEX-Record Adresse:%04x Zeile: %d\n", error, Adresse, lineNr);
           else
            {
             _cprintf("Fehler im Datenteil Zeile:%d!\n", lineNr);
             break;
            }
          }


///////////// checksumme ////////////////////
         if(j>(sizeof(LineBuffer)-2))
          {
           _cprintf("Pufferberlauf checksumme Zeile:%d!\n", lineNr);
           break;
          }
         else
          {
checkLastRecord:
           if((temp=hex2bin(&LineBuffer[j], 2))<0)
            {
             _cprintf("checksumme Zeichenfehler Zeile:%d!\n", lineNr);
             break;
            }
           else
            {
             if(byteHRecChecksumm)
              {
               _cprintf("checksumme falsch Zeile:%d!\n", lineNr);
               break;
              }
            }
          }
        }
      }
    } // while (!EOF)

   fclose(input);

   _cprintf("%d Zeilen gelesen!\n", lineNr);





//////////////////////////////////////////////////////////////////////////

// COM port (open and setup)

//////////////////////////////////////////////////////////////////////////

//   _getch(); // debugging

   if ((hPort =
      CreateFile( comInterface /*"COM1:"/*szPort*/, GENERIC_READ | GENERIC_WRITE,
                  0,                    // exclusive access
                  NULL,                 // no security attrs
                  OPEN_EXISTING,
                  FILE_ATTRIBUTE_NORMAL /*| FILE_FLAG_OVERLAPPED*/, // overlapped I/O
                  NULL )) == (HANDLE) -1 )
    {
     _cprintf("COM open "); // "error..."
     goto errorWait;
    }

// setup device buffers
   SetupComm( hPort, 4096, 4096 ) ;

// purge any information in the buffer
   PurgeComm( hPort, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR ) ;

// set up for none overlapped I/O

   SetCommTimeouts( hPort, &comTimeout); //CommTimeOuts ) ;
   wsprintf(LineBuffer, "%s9600,o,8,1", comInterface); // >> RTS/DTR_CONTROL_ENABLE // RTS und DTR statisch auf high
   if (!BuildCommDCB(LineBuffer/*"COM1:9600,o,8,2"*/ , &DeviceCon))  /*Xon Xoff Hardware-Flu deaktiviert*/
     goto closeOnErrorSetComm;

   DeviceCon.fBinary=1;
   DeviceCon.EvtChar=ETX;
   if(!SetCommState(hPort, &DeviceCon)) goto closeOnErrorSetComm;
//  EscapeCommFunction(hPort, SETDTR);      // wird schon beim ffnen gemacht
   _cprintf("\rsende Datei %s an %s ...",  argv[2], comInterface); // comInterface -> &argv[1]
   Sleep(100);
   EscapeCommFunction(hPort, CLRDTR);
   Sleep(200); // 100 waren zu knapp
   _cprintf("Connecting");
   i=3; // 1s
   do
    {
     _cprintf(".");
     if(writeCmd(cmdTimeoutOff, sizeof(cmdTimeoutOff))==sizeof(cmdTimeoutOff)) goto nextCmd;
    }while(--i);
error:
closeOnErrorSetComm:
   CloseHandle(hPort);
errorWait:
   _cprintf("error\nweiter mit Taste!\n");_getch(); while (!_kbhit());
   return;

nextCmd:
   _cprintf(".done\n"); // connection
   if(!writeCmd(cmdFill0,sizeof(cmdFill0))) goto error;

// **********************************************************************

// H   H   A   U   U PPPP  TTTTT  SSS   CCC H   H L    EEEE III FFFF EEEE
// H   H  A A  U   U P   P   T   S     C    H   H L    E     I  F    E
// HHHHH A   A U   U PPPP    T    SSS  C    HHHHH L    EEE   I  FFF  EEE
// H   H AAAAA U   U P       T       S C    H   H L    E     I  F    E
// H   H A   A  UUU  P       T    SSS   CCC H   H LLLL EEEE III F    EEEE

// **********************************************************************


///////////////////// write Blocks of Records ///////////////////////////

   i=0;
   while(i<= RAM_END)
    {
     if((unsigned)BufferSrc[i]<256)   // 1. Byte der Hex Sequenz gefunden ?
      {
       char * dst=&LineBuffer[4];

       j=i;
       *(long*)LineBuffer=SOH|0x3000|(long)(i+offs)<<16; // SOH '0' adr.l adr.h
    // LineBuffer[1] = adr<0x20000? (adr<0x10000? CMD_WBLK0:CMD_WBLK1):CMD_WBLK32;
       do
        {
         unsigned tmp=BufferSrc[j++];

         if(tmp>256) {j--; break;}

         *dst=tmp;
         dst++;
         if((j-i)>=64) break;
        }
       while(j <= RAM_END);             // Anzahl Bytes ermitteln

       *dst++=(temp=crc16((unsigned char*)LineBuffer, j-i+4))>>8;
       *dst=temp;
       if(!sendBlock(LineBuffer, j-i+6)) goto error;

//////////////////////////// user break /////////////////////////////////
       if(_kbhit())
        {
         while(_kbhit()) if(_getch()==27)    // ESC
          {
           _cprintf("Ausgabe abgebrochen!\n");
           goto finish;
          }
         else _cprintf("\007");              // Bell (output sound, handle ^C on error)
         _cprintf("\rESC zum Abbruch druecken\n");
        }
//////////////////////////////////////////////////////////////////////////

       i=j;
      }
     else  i++;
    }
////////////////////////// Ende Hauptschleife ////////////////////////////


   if(!writeCmd(cmdPowerDown,sizeof(cmdPowerDown))) goto error;

finish:
   CloseHandle(hPort);
  } // file was successfully opened (and is already closed)
 _cprintf("fertig!\n");
}

